# -*- coding: utf-8 -*-

#   ZX Spectrum Emulator.
#   https://github.com/kosarev/zx
#
#   Copyright (C) 2017-2019 Ivan Kosarev.
#   ivan@kosarev.info
#
#   Published under the MIT license.


import io
import zipfile
import zx
from ._data import ArchiveFileFormat


class ZIPFileFormat(ArchiveFileFormat):
    _NAME = 'ZIP'

    def read_files(self, image):
        file = io.BytesIO(image)
        with zipfile.ZipFile(file, 'r') as zf:
            for name in zf.namelist():
                with zf.open(name) as mf:
                    yield name, mf.read()



# -*- coding: utf-8 -*-

#   ZX Spectrum Emulator.
#   https://github.com/kosarev/zx
#
#   Copyright (C) 2017-2020 Ivan Kosarev.
#   ivan@kosarev.info
#
#   Published under the MIT license.

import time


class Time(object):
    def __init__(self):
        self._seconds = 0

    def get(self):
        return self._seconds

    def advance(self, s):
        self._seconds += s


def get_timestamp():
    # TODO: We can use this since Python 3.7.
    # return time.time_ns() / (10 ** 9)
    return time.time()


def get_elapsed_time(timestamp):
    return get_timestamp() - timestamp



# -*- coding: utf-8 -*-

#   ZX Spectrum Emulator.
#   https://github.com/kosarev/zx
#
#   Copyright (C) 2017-2020 Ivan Kosarev.
#   ivan@kosarev.info
#
#   Published under the MIT license.

import os
from ._data import ArchiveFileFormat
from ._error import Error
from ._rzx import RZXFileFormat
from ._scr import SCRFileFormat
from ._tap import TAPFileFormat
from ._tzx import TZXFileFormat
from ._wav import WAVFileFormat
from ._z80snapshot import Z80SnapshotFormat
from ._zip import ZIPFileFormat
from ._zxb import ZXBasicCompilerSourceFormat


def _open_file_or_url(path):
    if path.startswith(('http:', 'https:', 'ftp:')):
        import urllib.request
        import urllib.error
        try:
            HEADERS = {
                'User-Agent': 'Mozilla/5.0 (Windows; U; Windows NT 5.1; '
                              'en-US; rv:1.9.0.7) Gecko/2009021910 '
                              'Firefox/3.0.7',
            }
            req = urllib.request.Request(path, headers=HEADERS)
            return urllib.request.urlopen(req)
        except urllib.error.HTTPError as e:
            raise Error('Cannot read remote file: %s, code %d.' % (
                            e.reason, e.code))

    return open(path, 'rb')


def detect_file_format(image, filename_extension):
    KNOWN_FORMATS = [
        ('.zxb', None, ZXBasicCompilerSourceFormat),
        ('.rzx', b'RZX!', RZXFileFormat),
        ('.scr', None, SCRFileFormat),
        ('.tap', None, TAPFileFormat),
        ('.tzx', b'ZXTape!', TZXFileFormat),
        ('.wav', b'RIFF', WAVFileFormat),
        ('.z80', None, Z80SnapshotFormat),
        ('.zip', b'PK\x03\x04', ZIPFileFormat),
    ]

    filename_extension = filename_extension.lower()

    # First, try formats without signatures.
    for ext, signature, format in KNOWN_FORMATS:
        if not signature and filename_extension == ext:
            return format

    # Then, look at the signature.
    if image:
        for ext, signature, format in KNOWN_FORMATS:
            if signature and image[:len(signature)] == signature:
                return format

    # Finally, just try to guess by the given extension.
    for ext, signature, format in KNOWN_FORMATS:
        if filename_extension == ext:
            return format

    return None


def _parse_archive(format, image):
    candidates = []
    for member_name, member_image in format().read_files(image):
        base, ext = os.path.splitext(member_name)
        member_format = detect_file_format(member_image, ext)

        if not member_format:
            continue

        # Recursively parse member archives.
        if issubclass(member_format, ArchiveFileFormat):
            candidates.extend(_parse_archive(member_format, member_image))
            continue

        candidates.append((member_name, member_format, member_image))

    return candidates


def _parse_file_image(filename, image):
    base, ext = os.path.splitext(filename)
    format = detect_file_format(image, ext)
    if not format:
        raise Error('Cannot determine the format of file %r.' % filename)

    if issubclass(format, ArchiveFileFormat):
        candidates = _parse_archive(format, image)
        if len(candidates) == 0:
            raise Error('No files of known formats in archive %r.' %
                        filename)

        if len(candidates) > 1:
            raise Error(
                'More than one file of a known format in archive %r: %s.' % (
                    filename, ', '.join(repr(n) for n, f, im in candidates)))

        filename, format, image = candidates[0]

    return format().parse(filename, image)


def parse_file(filename):
    with _open_file_or_url(filename) as f:
        image = f.read()

    return _parse_file_image(filename, image)



# -*- coding: utf-8 -*-

#   ZX Spectrum Emulator.
#   https://github.com/kosarev/zx
#
#   Copyright (C) 2017-2019 Ivan Kosarev.
#   ivan@kosarev.info
#
#   Published under the MIT license.

import cairo
import gi
from ._device import Device
from ._device import DeviceEvent
from ._error import USER_ERRORS
from ._error import verbalize_error
from ._except import EmulationExit
from ._keyboard import KEYS_INFO
from ._time import get_elapsed_time
from ._time import get_timestamp
from ._utils import div_ceil
from ._z80snapshot import Z80SnapshotFormat
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gdk  # nopep8


SCREENCAST = False

PI = 3.1415926535


def rgb(color, alpha=1):
    assert color.startswith('#')
    assert len(color) == 7
    r = int(color[1:3], 16)
    g = int(color[3:5], 16)
    b = int(color[5:7], 16)
    return r / 0xff, g / 0xff, b / 0xff, alpha


def _draw_pause_sign(context, x, y, size, alpha):
    w = 0.1 * size
    h = 0.4 * size
    d = 0.15 * size
    context.rectangle(x - d, y - h / 2, w, h)
    context.rectangle(x + d - w, y - h / 2, w, h)
    context.fill()


def _draw_tape_sign(context, x, y, size, alpha, t=0):
    R = 0.10
    D = 0.33 - R
    H = 0.6
    RPM = 33.3

    context.set_line_width(size * 0.05)
    context.set_line_cap(cairo.LINE_CAP_ROUND)
    context.set_line_join(cairo.LINE_JOIN_ROUND)

    context.rectangle(x - size * 0.5, y - size * (H / 2), size, size * H)

    context.move_to(x - size * (D - 0.15), y - size * R)
    context.line_to(x + size * (D - 0.15), y - size * R)

    context.move_to(x - size * (D - R), y)
    context.new_sub_path()
    a = t * (RPM * 2 * PI / 60)
    context.arc(x - size * D, y, size * R, a, a + (2 * PI - 0.7))

    context.move_to(x + size * (D + R), y)
    context.new_sub_path()
    a += PI / 5
    # context.arc(x + size * D, y, size * R, 0, 2 * PI)
    context.arc(x + size * D, y, size * R, a, a + (2 * PI - 0.7))

    context.stroke()


def _draw_notification_circle(context, x, y, size, alpha):
    context.arc(x, y, size / 2, 0, 2 * PI)
    context.set_source_rgba(*rgb('#1e1e1e', alpha))
    context.fill()


def draw_pause_notification(context, x, y, size, alpha=1, t=0):
    _draw_notification_circle(context, x, y, size, alpha)

    context.set_source_rgba(*rgb('#ffffff', alpha))
    _draw_pause_sign(context, x, y, size, alpha)


def draw_tape_pause_notification(context, x, y, size, alpha=1, t=0):
    _draw_notification_circle(context, x, y, size, alpha)

    context.set_source_rgba(*rgb('#ffffff', alpha))
    _draw_tape_sign(context, x, y - size * 0.13, size * 0.5, alpha, t)
    _draw_pause_sign(context, x, y + size * 0.23, size * 0.5, alpha)


def draw_tape_resume_notification(context, x, y, size, alpha=1, t=0):
    _draw_notification_circle(context, x, y, size, alpha)

    context.set_source_rgba(*rgb('#ffffff', alpha))
    _draw_tape_sign(context, x, y - size * 0.015, size * 0.6, alpha, t)


class Notification(object):
    def __init__(self):
        self.clear()

    def set(self, draw, time):
        self._timestamp = get_timestamp()
        self._draw = draw
        self._time = time

    def clear(self):
        self._timestamp = None
        self._draw = None

    def draw(self, window_size, screen_size, context):
        if not self._timestamp:
            return

        width, height = screen_size
        window_width, window_height = window_size

        size = min(80, width * 0.2)
        x = (window_width - size) // 2
        y = (window_height - size) // 2

        alpha = 1.5 - get_elapsed_time(self._timestamp)
        alpha = max(0, min(0.7, alpha))

        if not alpha:
            self.clear()
            return

        self._draw(context, x + size / 2, y + size / 2, size, alpha,
                   self._time.get())


# TODO: A quick solution for making screencasts.
class Screencast(object):
    def __init__(self):
        self._counter = 0

    def on_draw(self, surface):
        if not SCREENCAST:
            return

        filename = '%05d.png' % self._counter
        surface.write_to_png(filename)
        self._counter += 1


class ScreenWindow(Device):
    _SCREEN_AREA_BACKGROUND_COLOUR = rgb('#1e1e1e')

    def __init__(self, emulator):
        super().__init__(emulator)

        self._queued_exception = None

        self._window = Gtk.Window()

        self._KEY_HANDLERS = {
            'ESCAPE': self._stop,
            'F10': self._stop,
            'F1': self._show_help,
            'F2': self._save_snapshot,
            'F3': self._choose_and_load_file,
            'F6': self.emulator._toggle_tape_pause,
            'F11': self._toggle_fullscreen,
            'PAUSE': self.emulator._toggle_pause,
        }

        self._EMULATOR_EVENT_HANDLERS = {
            DeviceEvent.PAUSE_STATE_UPDATED: self._on_updated_pause_state,
            DeviceEvent.QUANTUM_RUN: self._on_quantum_run,
            DeviceEvent.SCREEN_UPDATED: self._on_updated_screen,
            DeviceEvent.TAPE_STATE_UPDATED: self._on_updated_tape_state,
        }

        self._notification = Notification()
        self._screencast = Screencast()

        # TODO: Hide members like this.
        self.frame_width = 48 + 256 + 48
        self.frame_height = 48 + 192 + 40

        self.scale = 1 if SCREENCAST else 2

        self.area = Gtk.DrawingArea()
        self.area.connect('draw', self._on_draw_area)
        self._window.add(self.area)

        self._window.set_title('ZX Spectrum Emulator')
        if SCREENCAST:
            width, height = 640, 390
        else:
            width, height = (self.frame_width * self.scale,
                             self.frame_height * self.scale)
        self._window.resize(width, height)
        minimum_size = self.frame_width // 4, self.frame_height // 4
        self._window.set_size_request(*minimum_size)
        self._window.set_position(Gtk.WindowPosition.CENTER)
        self._window.connect('delete-event', self._on_done)

        self._window.show_all()

        self.frame_size = self.frame_width * self.frame_height
        self.frame = cairo.ImageSurface(cairo.FORMAT_RGB24,
                                        self.frame_width, self.frame_height)
        self.frame_data = self.frame.get_data()

        self.pattern = cairo.SurfacePattern(self.frame)
        if not SCREENCAST:
            self.pattern.set_filter(cairo.FILTER_NEAREST)

        self.keys = {'RETURN': KEYS_INFO['ENTER'],
                     'ALT_L': KEYS_INFO['CAPS SHIFT'],
                     'SHIFT_L': KEYS_INFO['CAPS SHIFT'],
                     'ALT_R': KEYS_INFO['SYMBOL SHIFT'],
                     'SHIFT_R': KEYS_INFO['SYMBOL SHIFT'],
                     'SPACE': KEYS_INFO['BREAK SPACE']}
        for id in ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
                   'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J',
                   'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T',
                   'U', 'V', 'W', 'X', 'Y', 'Z']:
            self.keys[id] = KEYS_INFO[id]

        self._window.connect('key-press-event', self._on_key_press)
        self._window.connect('key-release-event', self._on_key_release)
        self._window.connect('button-press-event', self._on_click)
        self._window.connect('window-state-event', self._on_window_state_event)

    def _on_draw_area(self, widget, context):
        window_size = self._window.get_size()
        window_width, window_height = window_size
        width = min(window_width,
                    div_ceil(window_height * self.frame_width,
                             self.frame_height))
        height = min(window_height,
                     div_ceil(window_width * self.frame_height,
                              self.frame_width))

        # Draw the background.
        context.save()
        context.rectangle(0, 0, window_width, window_height)
        context.set_source_rgba(*self._SCREEN_AREA_BACKGROUND_COLOUR)
        context.fill()

        # Draw the emulated screen.
        context.save()
        context.translate((window_width - width) // 2,
                          (window_height - height) // 2)
        context.scale(width / self.frame_width, height / self.frame_height)
        context.set_source(self.pattern)
        context.paint()
        context.restore()

        self._notification.draw(window_size, (width, height), context)
        context.restore()

        self._screencast.on_draw(context.get_group_target())

    def _on_updated_screen(self, pixels):
        self.frame_data[:] = pixels
        self.area.queue_draw()

    def _show_help(self):
        KEYS = [
            ('F1', 'Show help.'),
            ('F2', 'Save snapshot.'),
            ('F3', 'Load snapshot or tape file.'),
            ('F6', 'Pause/resume tape.'),
            ('F10', 'Quit.'),
            ('F11', 'Fullscreen/windowed mode.'),
            ('PAUSE', 'Pause/resume emulation.'),
        ]

        for entry in KEYS:
            print('%7s  %s' % entry)

    def _save_snapshot(self):
        # TODO: Add file filters.
        dialog = Gtk.FileChooserDialog(
            'Save snapshot', self._window,
            Gtk.FileChooserAction.SAVE,
            (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
             Gtk.STOCK_SAVE, Gtk.ResponseType.OK))
        dialog.set_do_overwrite_confirmation(True)
        if dialog.run() == Gtk.ResponseType.OK:
            try:
                self.emulator._save_snapshot_file(Z80SnapshotFormat,
                                                  dialog.get_filename())
            except USER_ERRORS as e:
                self._error_box('File error', verbalize_error(e))

        dialog.destroy()

    def _error_box(self, title, message):
        dialog = Gtk.MessageDialog(
            self._window, 0, Gtk.MessageType.ERROR,
            Gtk.ButtonsType.OK, title)
        dialog.format_secondary_text(message)
        dialog.run()
        dialog.destroy()

    def _choose_and_load_file(self):
        # TODO: Add file filters.
        dialog = Gtk.FileChooserDialog(
            'Load file', self._window,
            Gtk.FileChooserAction.OPEN,
            (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
             Gtk.STOCK_OPEN, Gtk.ResponseType.OK))

        if dialog.run() == Gtk.ResponseType.OK:
            filename = dialog.get_filename()
            try:
                self.emulator._load_file(filename)
            except USER_ERRORS as e:
                self._error_box('File error', verbalize_error(e))

        dialog.destroy()

    def _toggle_fullscreen(self):
        if self._is_fullscreen:
            self._window.unfullscreen()
        else:
            self._window.fullscreen()

    def _on_key(self, event, pressed):
        key_id = Gdk.keyval_name(event.keyval).upper()
        # print(key_id)

        if pressed and key_id in self._KEY_HANDLERS:
            self._KEY_HANDLERS[key_id]()

        key_info = self.keys.get(key_id, None)
        if key_info:
            self.emulator._pause(False)
            self.emulator._quit_playback_mode()
            self.emulator._handle_key_stroke(key_info, pressed)

    def _on_key_press(self, widget, event):
        self._on_key(event, pressed=True)

    def _on_key_release(self, widget, event):
        self._on_key(event, pressed=False)

    def _on_click(self, widget, event):
        if event.type == Gdk.EventType.BUTTON_PRESS:
            self.emulator._toggle_pause()
            return True
        elif event.type == Gdk.EventType._2BUTTON_PRESS:
            self._toggle_fullscreen()

    def _stop(self):
        self._queued_exception = EmulationExit()

    def _on_done(self, widget, context):
        self._stop()

    def _on_window_state_event(self, widget, event):
        state = event.new_window_state
        self._is_fullscreen = bool(state & Gdk.WindowState.FULLSCREEN)

    def _on_emulator_event(self, id, *args):
        self._EMULATOR_EVENT_HANDLERS[id](*args)

    def _on_updated_pause_state(self):
        if self.emulator._is_paused():
            self._notification.set(draw_pause_notification,
                                   self.emulator._emulation_time)
        else:
            self._notification.clear()

    def _on_updated_tape_state(self):
        tape_paused = self.emulator._is_tape_paused()
        draw = (draw_tape_pause_notification if tape_paused
                else draw_tape_resume_notification)
        tape_time = self.emulator._tape_player.get_time()
        self._notification.set(draw, tape_time)

    def _on_quantum_run(self):
        if self._queued_exception is not None:
            e = self._queued_exception
            self._queued_exception = None
            raise e

        while Gtk.events_pending():
            Gtk.main_iteration()

        self.area.queue_draw()

    def destroy(self):
        self._window.destroy()



# -*- coding: utf-8 -*-

#   ZX Spectrum Emulator.
#   https://github.com/kosarev/zx
#
#   Copyright (C) 2017-2020 Ivan Kosarev.
#   ivan@kosarev.info
#
#   Published under the MIT license.

from ._except import EmulatorException


class Error(EmulatorException):
    """Basic exception for the whole ZX module."""
    def __init__(self, reason, id=None):
        super().__init__(reason)

        if id:
            self.id = id


USER_ERRORS = Error, IOError


def verbalize_error(e):
    if isinstance(e, Error):
        reason, = e.args
        args = [reason]
    elif isinstance(e, IOError):
        code, reason = e.args
        args = ['%s (code %s).' % (reason, code)]
        if isinstance(e, FileNotFoundError):
            args.insert(0, e.filename)
    else:
        args = [type(e).__name__] + ['%s' % x for x in e.args]
    return ': '.join(args)



# -*- coding: utf-8 -*-

#   ZX Spectrum Emulator.
#   https://github.com/kosarev/zx
#
#   Copyright (C) 2017-2019 Ivan Kosarev.
#   ivan@kosarev.info
#
#   Published under the MIT license.


MASK16 = 0xffff


def div_ceil(a, b):
    return (a + b - 1) // b


def make16(hi, lo):
    return ((hi << 8) | lo) & MASK16


def _split16(nn):
    lo = nn & 0xff
    hi = (nn >> 8) & 0xff
    return lo, hi


def tupilize(x):
    if isinstance(x, (tuple, list)):
        return tuple(x)
    return x,



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#   ZX Spectrum Emulator.
#   https://github.com/kosarev/zx
#
#   Copyright (C) 2017-2019 Ivan Kosarev.
#   ivan@kosarev.info
#
#   Published under the MIT license.


import collections
import os
import sys
from ._data import ArchiveFileFormat
from ._data import SnapshotFormat
from ._data import SoundFileFormat
from ._emulator import Emulator
from ._error import Error
from ._error import USER_ERRORS
from ._error import verbalize_error
from ._except import EmulationExit
from ._file import detect_file_format
from ._file import parse_file
from ._rzx import make_rzx
from ._rzx import RZXFile


# Stores information about the running code.
class Profile(object):
    _annots = dict()

    def add_instr_addr(self, addr):
        self._annots[addr] = 'instr'

    def __iter__(self):
        for addr in sorted(self._annots):
            yield addr, self._annots[addr]


def pop_argument(args, error):
    if not args:
        raise Error(error)
    return args.pop(0)


def handle_extra_arguments(args):
    if args:
        raise Error('Extra argument %r.' % args[0])


def run(args):
    filename = None
    if args:
        filename = args.pop(0)
        handle_extra_arguments(args)

    with Emulator() as app:
        if filename:
            app._load_file(filename)
        app.run()


def profile(args):
    file_to_run = pop_argument(args, "The file to run is not specified.")
    profile_filename = pop_argument(args, "The profile filename is not "
                                          "specified.")
    handle_extra_arguments(args)

    profile = Profile()
    with Emulator(profile=profile) as app:
        app._load_file(file_to_run)
        app.run()

    # TODO: Amend profile data by reading them out first instead
    # of overwriting.
    with open(profile_filename, 'wt') as f:
        for addr, annot in profile:
            print('@0x%04x %s' % (addr, annot), file=f)


def dump(args):
    if not args:
        raise Error('The file to dump is not specified.')

    filename = args.pop(0)
    handle_extra_arguments(args)

    file = parse_file(filename)
    print(file)


def looks_like_filename(s):
    return '.' in s


def usage():
    print('Usage:')
    print('  zx [run] [<file>]')
    print('  zx [convert] <from-file> <to-filename>')
    print('  zx profile <file-to-run> <profile-filename>')
    print('  zx dump <file>')
    print('  zx help')
    sys.exit()


def test_file(filename):
    print('%r' % filename)

    def move(dest_dir):
        os.makedirs(dest_dir, exist_ok=True)

        # Make sure the destination filename is unique.
        dest_filename = filename
        while True:
            dest_path = os.path.join(dest_dir, dest_filename)
            if not os.path.exists(dest_path):
                break

            dest_filename, ext = os.path.splitext(dest_filename)
            dest_filename = dest_filename.rsplit('--', maxsplit=1)
            if len(dest_filename) == 1:
                dest_filename = dest_filename[0] + '--2'
            else:
                dest_filename = (dest_filename[0] + '--' +
                                 str(int(dest_filename[1]) + 1))

            dest_filename = dest_filename + ext

        os.rename(filename, dest_path)
        print('%r moved to %r' % (filename, dest_dir))

    with Emulator(speed_factor=None) as app:
        try:
            app._run_file(filename)
            move('passed')
        except Error as e:
            move(e.id)
        except:
            # TODO: Refine.
            return False

    return True


def test(args):
    for filename in args:
        if not test_file(filename):
            break


def fastforward(args):
    for filename in args:
        with Emulator(speed_factor=0) as app:
            app._run_file(filename)


def _convert_tape_to_snapshot(src, src_filename, src_format,
                              dest_filename, dest_format):
    assert issubclass(src_format, SoundFileFormat), src_format
    assert issubclass(dest_format, SnapshotFormat), dest_format

    with Emulator(speed_factor=None) as app:
        app.load_tape(src_filename)
        app._save_snapshot_file(dest_format, dest_filename)


def _convert_tape_to_tape(src, src_filename, src_format,
                          dest_filename, dest_format):
    assert issubclass(src_format, SoundFileFormat), src_format
    assert issubclass(dest_format, SoundFileFormat), dest_format

    dest_format().save_from_pulses(dest_filename, src.get_pulses())


def _convert_snapshot_to_snapshot(src, src_filename, src_format,
                                  dest_filename, dest_format):
    assert issubclass(src_format, SnapshotFormat), src_format
    assert issubclass(dest_format, SnapshotFormat), dest_format

    with Emulator(speed_factor=None) as app:
        app._load_file(src_filename)
        app._save_snapshot_file(dest_format, dest_filename)


def convert_file(src_filename, dest_filename):
    src = parse_file(src_filename)
    src_format = src.get_format()
    # print(src, '->', dest_filename)

    _, dest_ext = os.path.splitext(dest_filename)
    dest_format = detect_file_format(image=None, filename_extension=dest_ext)
    if not dest_format:
        raise Error('Cannot determine the format of file %r.' % (
                        dest_filename))

    CONVERTERS = [
        (SoundFileFormat, SoundFileFormat,
         _convert_tape_to_tape),
        (SoundFileFormat, SnapshotFormat,
         _convert_tape_to_snapshot),
        (SnapshotFormat, SnapshotFormat,
         _convert_snapshot_to_snapshot),
    ]

    for sf, df, conv in CONVERTERS:
        if issubclass(src_format, sf) and issubclass(dest_format, df):
            conv(src, src_filename, src_format, dest_filename, dest_format)
            return

    raise Error("Don't know how to convert from %s to %s files." % (
                src_format().get_name(),
                dest_format().get_name()))


def convert(args):
    if not args:
        raise Error('The file to convert from is not specified.')
    src_filename = args.pop(0)

    if not args:
        raise Error('The file to convert to is not specified.')
    dest_filename = args.pop(0)

    handle_extra_arguments(args)

    convert_file(src_filename, dest_filename)


def handle_command_line(args):
    # Guess the command by the arguments.
    if (not args or
            len(args) == 1 and looks_like_filename(args[0])):
        run(args)
        return

    if (len(args) == 2 and looks_like_filename(args[0]) and
            looks_like_filename(args[1])):
        convert(args)
        return

    # Handle an explicitly specified command.
    command = args[0]
    if command in ['help', '-help', '--help',
                   '-h', '-?',
                   '/h', '/help']:
        usage()
        return

    COMMANDS = {
        'convert': convert,
        'dump': dump,
        'profile': profile,
        'run': run,

        # TODO: Hidden commands for internal use.
        '__test': test,
        '__ff': fastforward,
    }

    if command not in COMMANDS:
        raise Error('Unknown command %r.' % command)

    COMMANDS[command](args[1:])


def main(args=None):
    if args is None:
        args = sys.argv[1:]

    try:
        handle_command_line(args)
    except EmulationExit:
        pass
    except USER_ERRORS as e:
        sys.exit('zx: %s' % verbalize_error(e))


if __name__ == "__main__":
    # import cProfile
    # cProfile.run('main()')
    main()



# -*- coding: utf-8 -*-

#   ZX Spectrum Emulator.
#   https://github.com/kosarev/zx
#
#   Copyright (C) 2017-2019 Ivan Kosarev.
#   ivan@kosarev.info
#
#   Published under the MIT license.


from ._time import Time


def _get_pilot_pulses(pilot_tone_len,
                      pilot_pulse_len=2168):
    for _ in range(pilot_tone_len):
        yield pilot_pulse_len, ('PILOT',)


def _get_sync_pulses(first_sync_pulse_len=667,
                     second_sync_pulse_len=735):
    yield first_sync_pulse_len, ('FIRST_SYNC_PULSE',)
    yield second_sync_pulse_len, ('SECOND_SYNC_PULSE',)


def _get_data_bits(data):
    for byte in data:
        for i in range(8):
            yield (byte & (1 << (7 - i))) != 0


def get_data_pulses(data,
                    zero_bit_pulse_len=855,
                    one_bit_pulse_len=1710):

    for bit in _get_data_bits(data):
        pulse = ((one_bit_pulse_len, ('ONE_BIT',)) if bit else
                 (zero_bit_pulse_len, ('ZERO_BIT',)))
        yield pulse
        yield pulse


def get_block_pulses(data,
                     pilot_pulse_len=2168,
                     first_sync_pulse_len=667,
                     second_sync_pulse_len=735,
                     zero_bit_pulse_len=855,
                     one_bit_pulse_len=1710,
                     pilot_tone_len=None):
    # Generate pilot tone.
    if pilot_tone_len is None:
        is_header = data[0] < 128
        pilot_tone_len = 8063 if is_header else 3223

    for pulse in _get_pilot_pulses(pilot_tone_len):
        yield pulse

    # Sync pulses.
    for pulse in _get_sync_pulses(first_sync_pulse_len,
                                  second_sync_pulse_len):
        yield pulse

    # Data.
    for pulse in get_data_pulses(data,
                                 zero_bit_pulse_len,
                                 one_bit_pulse_len):
        yield pulse


def tag_last_pulse(pulses):
    current_pulse = None
    for pulse in pulses:
        if current_pulse:
            yield current_pulse
        current_pulse = pulse

    if current_pulse:
        level, pulse, ids = current_pulse
        if 'END' not in ids:
            ids = tuple(list(ids) + ['END'])
            yield level, pulse, ids


class TapePlayer(object):
    def __init__(self):
        self._is_paused = True
        self._pulses = None
        self._tick = 0
        self._level = False
        self._pulse = 0
        self._ticks_per_frame = 69888  # TODO
        self._time = Time()

    def is_paused(self):
        return self._is_paused

    def pause(self, is_paused=True):
        self._is_paused = is_paused

    def unpause(self):
        self.pause(is_paused=False)

    def toggle_pause(self):
        self.pause(not self.is_paused())

    def is_end(self):
        return self._pulses is None

    def get_time(self):
        return self._time

    def load_parsed_file(self, file):
        self._pulses = file.get_pulses()
        self._level = False
        self.pause()

    def load_tape(self, file):
        self.load_parsed_file(file)

    def get_level_at_frame_tick(self, tick):
        assert self._tick <= tick, (self._tick, tick)

        while self._tick < tick:
            if self._is_paused:
                self._tick = tick
                continue

            # See if we already have a non-zero-length pulse.
            if self._pulse:
                ticks_to_skip = min(self._pulse, tick - self._tick)
                self._pulse -= ticks_to_skip
                self._tick += ticks_to_skip
                self._time.advance(ticks_to_skip /
                                   (self._ticks_per_frame * 50))
                continue

            # Get subsequent pulse, if any.
            new_pulse = None
            if self._pulses:
                for new_pulse in self._pulses:
                    break

            if new_pulse:
                # print(new_pulse)
                self._level, self._pulse, ids = new_pulse

                # The tape shall be considered stopped as soon as the last
                # pulse is fetched, and not on the next attempt to fetch a
                # pulse.
                if 'END' in ids:
                    self._pulses = None

                continue

            # Do nothing, if there are no more pulses available.
            self._pulses = None
            self._level = False
            self._tick = tick

        return self._level

    def skip_rest_of_frame(self):
        if self._tick < self._ticks_per_frame:
            self.get_level_at_frame_tick(self._ticks_per_frame)

        assert self._tick >= self._ticks_per_frame
        self._tick -= self._ticks_per_frame



# -*- coding: utf-8 -*-

#   ZX Spectrum Emulator.
#   https://github.com/kosarev/zx
#
#   Copyright (C) 2017-2019 Ivan Kosarev.
#   ivan@kosarev.info
#
#   Published under the MIT license.

import os
import tempfile
from ._data import File
from ._data import FileFormat
from ._error import Error


class ZXBasicCompilerProgram(File):
    pass


class ZXBasicCompilerSourceFormat(FileFormat):
    _NAME = 'ZXB'

    def parse(self, filename, image):
        try:
            import zxb
        except ModuleNotFoundError:
            raise Error('The ZX Basic compiler does not seem to be installed.')

        fields = {}

        class Emitter(zxb.CodeEmitter):
            def emit(self, **args):
                fields.update(args)

        with tempfile.TemporaryDirectory() as dir:
            path = os.path.join(dir, filename)
            with open(path, 'wb') as f:
                f.write(image)

            status = zxb.main(args=[path], emitter=Emitter())
            if status:
                raise Error('ZX Basic compiler returned %d.' % status)

        fields['program_bytes'] = bytes(fields['program_bytes'])

        return ZXBasicCompilerProgram(ZXBasicCompilerSourceFormat, fields)



# -*- coding: utf-8 -*-

#   ZX Spectrum Emulator.
#   https://github.com/kosarev/zx
#
#   Copyright (C) 2017-2019 Ivan Kosarev.
#   ivan@kosarev.info
#
#   Published under the MIT license.


from ._binary import BinaryParser
from ._data import SoundFile
from ._data import SoundFileFormat
from ._error import Error
from ._tape import get_block_pulses, get_data_pulses, tag_last_pulse


class TZXFile(SoundFile):
    _TICKS_FREQ = 3500000

    def __init__(self, fields):
        SoundFile.__init__(self, TZXFileFormat, fields)

    def _generate_pulses(self):
        level = False
        for block in self['blocks']:
            id = block['id']
            if id in ['0x10 (Standard Speed Data Block)',
                      '0x11 (Turbo Speed Data Block)',
                      '0x14 (Pure Data Block)']:
                # The block itself.
                data = block['data']
                if id == '0x10 (Standard Speed Data Block)':
                    pulses = get_block_pulses(data)
                elif id == '0x11 (Turbo Speed Data Block)':
                    pulses = get_block_pulses(
                        data=data,
                        pilot_pulse_len=block['pilot_pulse_len'],
                        first_sync_pulse_len=block['first_sync_pulse_len'],
                        second_sync_pulse_len=block['second_sync_pulse_len'],
                        zero_bit_pulse_len=block['zero_bit_pulse_len'],
                        one_bit_pulse_len=block['one_bit_pulse_len'],
                        pilot_tone_len=block['pilot_tone_len'])
                elif id == '0x14 (Pure Data Block)':
                    pulses = get_data_pulses(
                        data=data,
                        zero_bit_pulse_len=block['zero_bit_pulse_len'],
                        one_bit_pulse_len=block['one_bit_pulse_len'])
                else:
                    assert 0, id

                for pulse, ids in pulses:
                    yield level, pulse, ids
                    level = not level

                # Pause.
                pause_duration = block['pause_after_block_in_ms']

                # Pause blocks of zero duration shall be ignored,
                # and then the output level remains the same.
                if pause_duration == 0:
                    continue

                # At the end of the non-zero-duration pause the
                # output level shall be low.
                ''' TODO: Despite the specification, this cases
                          tape loading errors.
                if level:
                    # Give the high pulse 1ms of time and drop it.
                    yield level, self._TICKS_FREQ // 1000, ('PAUSE',)
                    pause_duration -= 1
                    level = not level

                assert not level
                '''

                if pause_duration:
                    yield (level, pause_duration * self._TICKS_FREQ // 1000,
                           ('PAUSE',))
            elif id == '0x30 (Text Description)':
                print(block['text'])
            elif id in ['0x32 (Archive Info)', '0x21 (Group Start)',
                        '0x22 (Group End)']:
                pass
            else:
                assert 0, block  # TODO

    def get_pulses(self):
        return tag_last_pulse(self._generate_pulses())


class TZXFileFormat(SoundFileFormat):
    _NAME = 'TZX'

    def _parse_standard_speed_data_block(self, parser):
        block = parser.parse([('pause_after_block_in_ms', '<H'),
                              ('data_size', '<H')])
        block.update({'id': '0x10 (Standard Speed Data Block)',
                      'data': parser.extract_block(block['data_size'])})
        del block['data_size']
        return block

    def _parse_turbo_speed_data_block(self, parser):
        block = parser.parse([('pilot_pulse_len', '<H'),
                              ('first_sync_pulse_len', '<H'),
                              ('second_sync_pulse_len', '<H'),
                              ('zero_bit_pulse_len', '<H'),
                              ('one_bit_pulse_len', '<H'),
                              ('pilot_tone_len', '<H'),
                              ('used_bits_in_last_byte', '<B'),
                              ('pause_after_block_in_ms', '<H'),
                              ('data_size_lo', '<H'),
                              ('data_size_hi', '<B')])
        data_size = (block.pop('data_size_hi') * 0x10000 +
                     block.pop('data_size_lo'))
        data_size_in_bits = data_size * 8 + block.pop('used_bits_in_last_byte')
        block.update({'id': '0x11 (Turbo Speed Data Block)',
                      'data_size_in_bits': data_size_in_bits,
                      'data': parser.extract_block(data_size)})
        return block

    def _parse_pure_data_block(self, parser):
        block = parser.parse([('zero_bit_pulse_len', '<H'),
                              ('one_bit_pulse_len', '<H'),
                              ('used_bits_in_last_byte', '<B'),
                              ('pause_after_block_in_ms', '<H'),
                              ('data_size_lo', '<H'),
                              ('data_size_hi', '<B')])
        data_size = (block.pop('data_size_hi') * 0x10000 +
                     block.pop('data_size_lo'))
        data_size_in_bits = data_size * 8 + block.pop('used_bits_in_last_byte')
        block.update({'id': '0x14 (Pure Data Block)',
                      'data_size_in_bits': data_size_in_bits,
                      'data': parser.extract_block(data_size)})
        return block

    def _parse_group_start(self, parser):
        length = parser.parse_field('B', 'name_length')
        name = parser.extract_block(length)
        # print('Group start: %r.' % name)
        return {'id': '0x21 (Group Start)',
                'name': name}

    def _parse_group_end(self, parser):
        # TODO: Check for a matching group start?
        # print('Group end.')
        return {'id': '0x22 (Group End)'}

    def _parse_text_description(self, parser):
        size = parser.parse_field('B', 'text_size')
        text = parser.extract_block(size)
        return {'id': '0x30 (Text Description)',
                'text': text}

    _ARCHIVE_INFO_STRING_IDS = {
        0x00: 'Full title',
        0x01: 'Software house/publisher',
        0x02: 'Author(s)',
        0x03: 'Year of publication',
        0x04: 'Language',
        0x05: 'Game/utility type',
        0x06: 'Price',
        0x07: 'Protection scheme/loader',
        0x08: 'Origin',
        0xFF: 'Comment(s)',
    }

    def _parse_archive_info(self, parser):
        block_size = parser.parse_field('<H', 'block_size')
        num_of_strings = parser.parse_field('B', 'num_of_strings')
        for _ in range(num_of_strings):
            id = parser.parse_field('B', 'id')
            length = parser.parse_field('B', 'length')
            body = parser.extract_block(length)

            if id not in self._ARCHIVE_INFO_STRING_IDS:
                raise Error('Unknown TZX archive info string id 0x%02x.' % id)

            # print('%s: %s' % (self._ARCHIVE_INFO_STRING_IDS[id], body))
        # TODO: Encode all the details.
        return {'id': '0x32 (Archive Info)'}

    _BLOCK_PARSERS = {
        0x10: _parse_standard_speed_data_block,
        0x11: _parse_turbo_speed_data_block,
        0x14: _parse_pure_data_block,
        0x21: _parse_group_start,
        0x22: _parse_group_end,
        0x30: _parse_text_description,
        0x32: _parse_archive_info,
    }

    def _parse_block(self, parser):
        block_id = parser.parse_field('B', 'block_id')
        if block_id not in self._BLOCK_PARSERS:
            raise Error('Unsupported TZX block id 0x%x.' % block_id)

        return self._BLOCK_PARSERS[block_id](self, parser)

    def parse(self, filename, image):
        parser = BinaryParser(image)

        # Parse header.
        header = parser.parse([('signature', '8s'),
                               ('major_revision', 'B'),
                               ('minor_revision', 'B')])
        tzx_signature = b'ZXTape!\x1a'
        if header['signature'] != tzx_signature:
            raise Error('Bad TZX file signature %r; expected %r.' % (
                        header['signature'], tzx_signature))

        # Parse blocks.
        blocks = []
        while not parser.is_eof():
            blocks.append(self._parse_block(parser))

        return TZXFile({**header, 'blocks': blocks})



# -*- coding: utf-8 -*-

#   ZX Spectrum Emulator.
#   https://github.com/kosarev/zx
#
#   Copyright (C) 2017-2020 Ivan Kosarev.
#   ivan@kosarev.info
#
#   Published under the MIT license.

import enum


class DeviceEvent(enum.Enum):
    PAUSE_STATE_UPDATED = enum.auto()
    QUANTUM_RUN = enum.auto()
    SCREEN_UPDATED = enum.auto()
    TAPE_STATE_UPDATED = enum.auto()


class Device(object):
    def __init__(self, emulator):
        self.emulator = emulator



# -*- coding: utf-8 -*-

#   ZX Spectrum Emulator.
#   https://github.com/kosarev/zx
#
#   Copyright (C) 2017-2019 Ivan Kosarev.
#   ivan@kosarev.info
#
#   Published under the MIT license.

from ._utils import tupilize


class KeyInfo(object):
    def __init__(self, id, index):
        self.ID = id
        self.INDEX = index  # Left to right, then top to bottom.
        halfrow_index = index // 5
        index_in_halfrow = index % 5
        is_leftside = halfrow_index % 2 == 0

        # Compute port address lines and bit positions.
        if is_leftside:
            self.ADDRESS_LINE = 11 - halfrow_index // 2
            self.PORT_BIT = index_in_halfrow
        else:
            self.ADDRESS_LINE = halfrow_index // 2 + 12
            self.PORT_BIT = 4 - index_in_halfrow


_KEY_IDS = [
    '1', '2', '3', '4', '5', '6', '7', '8', '9', '0',
    'Q', 'W', 'E', 'R', 'T', 'Y', 'U', 'I', 'O', 'P',
    'A', 'S', 'D', 'F', 'G', 'H', 'J', 'K', 'L', 'ENTER',
    ('CAPS SHIFT', 'CS'), 'Z', 'X', 'C', 'V',
    'B', 'N', 'M', ('SYMBOL SHIFT', 'SS'), ('BREAK SPACE', 'SPACE')]

KEYS_INFO = dict()
for index, ids in enumerate(_KEY_IDS):
    ids = tupilize(ids)
    id, *aliases = ids
    info = KeyInfo(id, index)
    for i in ids:
        KEYS_INFO[i] = info


class KeyboardState(object):
    _state = [0xff] * 8

    def read_port(self, addr):
        n = 0xff
        addr ^= 0xffff

        if addr & (1 << 8):
            n &= self._state[0]
        if addr & (1 << 9):
            n &= self._state[1]
        if addr & (1 << 10):
            n &= self._state[2]
        if addr & (1 << 11):
            n &= self._state[3]
        if addr & (1 << 12):
            n &= self._state[4]
        if addr & (1 << 13):
            n &= self._state[5]
        if addr & (1 << 14):
            n &= self._state[6]
        if addr & (1 << 15):
            n &= self._state[7]

        return n

    def handle_key_stroke(self, key_info, pressed):
        # print(key_info.id)
        addr_line = key_info.ADDRESS_LINE
        mask = 1 << key_info.PORT_BIT

        if pressed:
            self._state[addr_line - 8] &= mask ^ 0xff
        else:
            self._state[addr_line - 8] |= mask



# -*- coding: utf-8 -*-

#   ZX Spectrum Emulator.
#   https://github.com/kosarev/zx
#
#   Copyright (C) 2017-2019 Ivan Kosarev.
#   ivan@kosarev.info
#
#   Published under the MIT license.

import zx


if __name__ == '__main__':
    zx.main()



# -*- coding: utf-8 -*-

#   ZX Spectrum Emulator.
#   https://github.com/kosarev/zx
#
#   Copyright (C) 2017-2020 Ivan Kosarev.
#   ivan@kosarev.info
#
#   Published under the MIT license.


import collections
from ._binary import BinaryParser, BinaryWriter
from ._data import MachineSnapshot
from ._data import ProcessorSnapshot
from ._data import SnapshotFormat
from ._data import UnifiedSnapshot
from ._utils import _split16


class _SCRSnapshot(MachineSnapshot):
    def get_unified_snapshot(self):
        # The address of the endless loop.
        LOOP_ADDR = 0x8000

        processor_fields = {
            'pc': LOOP_ADDR,
            'iff1': 0,
            'iff2': 0,
        }

        fields = {
            'processor_snapshot': ProcessorSnapshot(processor_fields),
            'border_color': 0,
        }

        memory_blocks = fields.setdefault('memory_blocks', [])
        memory_blocks.append((0x4000, self['dot_patterns']))
        memory_blocks.append((0x4000 + 6144, self['colour_attrs']))

        # LOOP_ADDR: jp LOOP_ADDR
        memory_blocks.append((LOOP_ADDR, b'\xc3' + bytes(_split16(LOOP_ADDR))))

        return UnifiedSnapshot(SCRFileFormat, fields)

    def get_file_image(self):
        return self['dot_patterns'] + self['colour_attrs']


class SCRFileFormat(SnapshotFormat):
    _NAME = 'SCR'

    _FIELDS = ['6144s:dot_patterns', '768s:colour_attrs']

    def parse(self, filename, image):
        parser = BinaryParser(image)
        fields = collections.OrderedDict(id='scr_snapshot')
        fields.update(parser.parse(self._FIELDS))
        return _SCRSnapshot(SCRFileFormat, fields)

    def make_snapshot(self, state):
        screen = state.get_memory_block(0x4000, 6 * 1024 + 768)
        return self.parse(screen)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#   ZX Spectrum Emulator.
#   https://github.com/kosarev/zx
#
#   Copyright (C) 2017-2020 Ivan Kosarev.
#   ivan@kosarev.info
#
#   Published under the MIT license.

import time
from ._data import MachineSnapshot
from ._data import SoundFile
from ._device import DeviceEvent
from ._error import Error
from ._except import EmulationExit
from ._file import parse_file
from ._gui import ScreenWindow
from ._keyboard import KeyboardState
from ._keyboard import KEYS_INFO
from ._machine import RunEvents
from ._machine import Spectrum48
from ._rzx import RZXFile
from ._tape import TapePlayer
from ._time import Time
from ._z80snapshot import Z80SnapshotFormat
from ._zxb import ZXBasicCompilerProgram


# TODO: Rework to a time machine interface.
class PlaybackPlayer(object):
    def __init__(self, file):
        assert isinstance(file, RZXFile)
        self._recording = file

    def find_recording_info_chunk(self):
        for chunk in self._recording['chunks']:
            if chunk['id'] == 'info':
                return chunk
        assert 0  # TODO

    def get_chunks(self):
        return self._recording['chunks']


class Emulator(object):
    _SPIN_V0P5_INFO = {'id': 'info',
                       'creator': b'SPIN 0.5            ',
                       'creator_major_version': 0,
                       'creator_minor_version': 5}

    def __init__(self, speed_factor=1.0, profile=None, devices=None):
        # TODO: Double-underscore or make public.
        self._emulation_time = Time()
        self.__speed_factor = speed_factor

        self.__is_paused_flag = False
        self.__events_to_signal = RunEvents.NO_EVENTS

        # Don't even create the window on full throttle.
        self.__devices = devices if devices is not None else []
        if devices is None and self.__speed_factor is not None:
            self.__devices = [ScreenWindow(self)]

        self.__machine = Spectrum48()

        self.__keyboard_state = KeyboardState()
        self.__machine.set_on_input_callback(self.__on_input)

        # TODO: Double-underscore or make public.
        self._tape_player = TapePlayer()

        self.__playback_player = None
        self.__playback_samples = None

        self.__profile = profile
        if self.__profile:
            self.__machine.set_breakpoints(0, 0x10000)

    def __enter__(self):
        return self

    def destroy(self):
        for device in self.__devices:
            device.destroy()

    def __exit__(self, type, value, tb):
        self.destroy()

    # TODO: Double-underscore or make public.
    def _is_paused(self):
        return self.__is_paused_flag

    def __notify(self, id, *args):
        for device in self.__devices:
            device._on_emulator_event(id, *args)

    # TODO: Double-underscore or make public.
    def _pause(self, is_paused=True):
        self.__is_paused_flag = is_paused
        self.__notify(DeviceEvent.PAUSE_STATE_UPDATED)

    # TODO: Double-underscore or make public.
    def _toggle_pause(self):
        self._pause(not self._is_paused())

    # TODO: Double-underscore or make public.
    def _save_snapshot_file(self, format, filename):
        with open(filename, 'wb') as f:
            snapshot = format().make_snapshot(self.__machine)
            # TODO: make_snapshot() shall always return a snapshot object.
            if issubclass(type(snapshot), MachineSnapshot):
                image = snapshot.get_file_image()
            else:
                image = snapshot
            f.write(image)

    def stop(self):
        raise EmulationExit()

    # TODO: Double-underscore or make public.
    def _is_tape_paused(self):
        return self._tape_player.is_paused()

    def __pause_tape(self, is_paused=True):
        self._tape_player.pause(is_paused)
        self.__notify(DeviceEvent.TAPE_STATE_UPDATED)

    def __unpause_tape(self):
        self.__pause_tape(is_paused=False)

    def _toggle_tape_pause(self):
        self.__pause_tape(not self._is_tape_paused())

    def __load_tape_to_player(self, file):
        self._tape_player.load_tape(file)
        self.__pause_tape()

    def __is_end_of_tape(self):
        return self._tape_player.is_end()

    # TODO: Double-underscore or make public.
    def _handle_key_stroke(self, key_info, pressed):
        self.__keyboard_state.handle_key_stroke(key_info, pressed)

    def __translate_key_strokes(self, keys):
        for key in keys:
            if isinstance(key, int):
                yield from str(key)
            else:
                yield key

    def __generate_key_strokes(self, *keys):
        for key in self.__translate_key_strokes(keys):
            strokes = key.split('+')
            # print(strokes)

            for id in strokes:
                # print(id)
                self._handle_key_stroke(KEYS_INFO[id], pressed=True)
                self.__run(0.05)

            for id in reversed(strokes):
                # print(id)
                self._handle_key_stroke(KEYS_INFO[id], pressed=False)
                self.__run(0.05)

    def __on_input(self, addr):
        # Handle playbacks.
        if self.__playback_samples:
            sample = None
            for sample in self.__playback_samples:
                break

            if sample == 'END_OF_FRAME':
                raise Error(
                    'Too few input samples at frame %d of %d. '
                    'Given %d, used %d.' % (
                        self.playback_frame_count,
                        len(self.playback_chunk['frames']),
                        len(self.__playback_samples), sample_i),
                    id='too_few_input_samples')

            # print('__on_input() returns %d' % sample)
            return sample

        # Scan keyboard.
        n = 0xbf
        n &= self.__keyboard_state.read_port(addr)

        # TODO: Use the tick when the ear value is sampled
        #       instead of the tick of the beginning of the input
        #       cycle.
        tick = self.__machine.get_ticks_since_int()
        if self._tape_player.get_level_at_frame_tick(tick):
            n |= 0x40

        END_OF_TAPE = RunEvents.END_OF_TAPE
        if END_OF_TAPE in self.__events_to_signal and self.__is_end_of_tape():
            self.__machine.raise_events(END_OF_TAPE)
            self.__events_to_signal &= ~END_OF_TAPE

        # print('0x%04x 0x%02x' % (addr, n))

        return n

    def __save_crash_rzx(self, player, state, chunk_i, frame_i):
        snapshot = Z80SnapshotFormat().make(state)

        crash_recording = {
            'id': 'input_recording',
            'chunks': [
                player.find_recording_info_chunk(),
                {
                    'id': 'snapshot',
                    'image': snapshot,
                },
                {
                    'id': 'port_samples',
                    'first_tick': 0,
                    'frames': recording['chunks'][chunk_i]['frames'][frame_i:],
                },
            ],
        }

        with open('__crash.z80', 'wb') as f:
            f.write(snapshot)

        with open('__crash.rzx', 'wb') as f:
            f.write(make_rzx(crash_recording))

    def __enter_playback_mode(self):
        # Interrupts are supposed to be controlled by the
        # recording.
        self.__machine.suppress_int()
        self.__machine.allow_int_after_ei()
        # self.__machine.enable_trace()

    # TODO: Double-underscore or make public.
    def _quit_playback_mode(self):
        self.__playback_player = None
        self.__playback_samples = None

        self.__machine.suppress_int(False)
        self.__machine.allow_int_after_ei(False)

    def __get_playback_samples(self):
        # TODO: Have a class describing playback state.
        self.playback_frame_count = 0
        self.playback_chunk = 0
        self.playback_sample_values = []
        self.playback_sample_i = 0

        frame_count = 0
        for chunk_i, chunk in enumerate(self.__playback_player.get_chunks()):
            if isinstance(chunk, MachineSnapshot):
                self.__machine.install_snapshot(chunk)
                continue

            if chunk['id'] != 'port_samples':
                continue

            self.__machine.set_ticks_since_int(chunk['first_tick'])

            for frame_i, frame in enumerate(chunk['frames']):
                num_of_fetches, samples = frame
                # print(num_of_fetches, samples)

                self.__machine.set_fetches_limit(num_of_fetches)
                # print(num_of_fetches, samples, flush=True)

                # print('START_OF_FRAME', flush=True)
                yield 'START_OF_FRAME'

                for sample_i, sample in enumerate(samples):
                    # print(self.__machine.get_fetches_limit())
                    # fetch = num_of_fetches -
                    #         self.__machine.get_fetches_limit()
                    # print('Input at fetch', fetch, 'of', num_of_fetches)
                    # TODO: print('read_port 0x%04x 0x%02x' % (addr, n),
                    #             flush=True)

                    # TODO: Have a class describing playback state.
                    self.playback_frame_count = frame_count
                    self.playback_chunk = chunk
                    self.playback_sample_values = samples
                    self.playback_sample_i = sample_i
                    # print(frame_count, chunk_i, frame_i, sample_i, sample,
                    #       flush=True)

                    yield sample

                # print('END_OF_FRAME', flush=True)
                yield 'END_OF_FRAME'

                frame_count += 1

    def __run_quantum(self):
        if self.__playback_player:
            creator_info = self.__playback_player.find_recording_info_chunk()

        if True:  # TODO
            self.__notify(DeviceEvent.QUANTUM_RUN)

            # TODO: For debug purposes.
            '''
            frame_count += 1
            if frame_count == -12820:
                frame_state = self.__machine.clone()
                self.__save_crash_rzx(player, frame_state, chunk_i, frame_i)
                assert 0

            if frame_count == -65952 - 1000:
                self.__machine.enable_trace()
            '''

            if self._is_paused():
                # Give the CPU some spare time.
                if self.__speed_factor:
                    time.sleep((1 / 50) * self.__speed_factor)
                return

            events = RunEvents(self.__machine.run())
            # TODO: print(events)

            if RunEvents.BREAKPOINT_HIT in events:
                self.on_breakpoint()

                if self.__profile:
                    pc = self.__machine.get_pc()
                    self.__profile.add_instr_addr(pc)

                # SPIN v0.5 skips executing instructions
                # of the bytes-saving ROM procedure in
                # fast save mode.
                if (self.__playback_samples and
                        creator_info == self._SPIN_V0P5_INFO and
                        self.__machine.get_pc() == 0x04d4):
                    sp = self.__machine.get_sp()
                    ret_addr = self.__machine.read16(sp)
                    self.__machine.set_sp(sp + 2)
                    self.__machine.set_pc(ret_addr)

            if RunEvents.END_OF_FRAME in events:
                self.__machine.render_screen()

                pixels = self.__machine.get_frame_pixels()
                self.__notify(DeviceEvent.SCREEN_UPDATED, pixels)

                self._tape_player.skip_rest_of_frame()
                self._emulation_time.advance(1 / 50)

                if self.__speed_factor:
                    time.sleep((1 / 50) * self.__speed_factor)

            if (self.__playback_samples and
                    RunEvents.FETCHES_LIMIT_HIT in events):
                # Some simulators, e.g., SPIN, may store an interrupt
                # point in the middle of a IX- or IY-prefixed
                # instruction, so we continue until such
                # instruction, if any, is completed.
                if self.__machine.get_iregp_kind() != 'hl':
                    self.__machine.set_fetches_limit(1)
                    return

                # SPIN doesn't update the fetch counter if the last
                # instruction in frame is IN.
                if (self.__playback_samples and
                        creator_info == self._SPIN_V0P5_INFO and
                        self.playback_sample_i + 1 <
                        len(self.playback_sample_values)):
                    self.__machine.set_fetches_limit(1)
                    return

                sample = None
                for sample in self.__playback_samples:
                    break
                if sample != 'END_OF_FRAME':
                    raise Error(
                        'Too many input samples at frame %d of %d. '
                        'Given %d, used %d.' % (
                            self.playback_frame_count,
                            len(self.playback_chunk['frames']),
                            len(self.__playback_samples),
                            self.playback_sample_i + 1),
                        id='too_many_input_samples')

                sample = None
                for sample in self.__playback_samples:
                    break
                if sample is None:
                    self.stop()
                    return

                assert sample == 'START_OF_FRAME'
                self.__machine.on_handle_active_int()

    def __run(self, duration):
        end_time = self._emulation_time.get() + duration
        while self._emulation_time.get() < end_time:
            self.__run_quantum()

    def run(self):
        # TODO: Remove this 'try', and let the exception to
        # propagate further.
        try:
            while True:
                self.__run_quantum()

            self._quit_playback_mode()
        except EmulationExit:
            pass

    def __load_input_recording(self, file):
        self.__playback_player = PlaybackPlayer(file)
        creator_info = self.__playback_player.find_recording_info_chunk()

        # SPIN v0.5 alters ROM to implement fast tape loading,
        # but that affects recorded RZX files.
        if creator_info == self._SPIN_V0P5_INFO:
            self.__machine.set_memory_block(0x1f47, b'\xf5')

        # The bytes-saving ROM procedure needs special processing.
        self.__machine.set_breakpoint(0x04d4)

        # Process frames in order.
        self.__playback_samples = self.__get_playback_samples()
        sample = None
        for sample in self.__playback_samples:
            break
        assert sample == 'START_OF_FRAME'

    def __reset_and_wait(self):
        self.__machine.set_pc(0x0000)
        self.__run(1.8)

    def __load_zx_basic_compiler_program(self, file):
        assert isinstance(file, ZXBasicCompilerProgram)

        self.__reset_and_wait()

        # CLEAR <entry_point>
        entry_point = file['entry_point']
        self.__generate_key_strokes('X', entry_point, 'ENTER')

        self.__machine.set_memory_block(entry_point, file['program_bytes'])

        # RANDOMIZE USR <entry_point>
        self.__generate_key_strokes('T', 'CS+SS', 'L', entry_point, 'ENTER')

        # assert 0, list(file)

    # TODO: Double-underscore or make public.
    def _load_file(self, filename):
        file = parse_file(filename)

        if isinstance(file, MachineSnapshot):
            self.__machine.install_snapshot(file)
        elif isinstance(file, RZXFile):
            self.__load_input_recording(file)
            self.__enter_playback_mode()
        elif isinstance(file, SoundFile):
            self.__load_tape_to_player(file)
        elif isinstance(file, ZXBasicCompilerProgram):
            self.__load_zx_basic_compiler_program(file)
        else:
            raise Error("Don't know how to load file %r." % filename)

    # TODO: Double-underscore or make public.
    def _run_file(self, filename):
        self._load_file(filename)
        self.run()

    def load_tape(self, filename):
        tape = parse_file(filename)
        if not isinstance(tape, SoundFile):
            raise Error('%r does not seem to be a tape file.' % filename)

        # Let the initialization complete.
        self.__reset_and_wait()

        # Type in 'LOAD ""'.
        self.__generate_key_strokes('J', 'SS+P', 'SS+P', 'ENTER')

        # Load and run the tape.
        self.__load_tape_to_player(tape)
        self.__unpause_tape()

        # Wait till the end of the tape.
        self.__events_to_signal |= RunEvents.END_OF_TAPE
        while not self.__is_end_of_tape():
            self.__run_quantum()

    def set_breakpoint(self, addr):
        self.__machine.set_breakpoint(addr)

    def on_breakpoint(self):
        pass

    def get_memory_view(self, addr, size):
        return self.__machine.get_memory_block(addr, size)



# -*- coding: utf-8 -*-

#   ZX Spectrum Emulator.
#   https://github.com/kosarev/zx
#
#   Copyright (C) 2017-2020 Ivan Kosarev.
#   ivan@kosarev.info
#
#   Published under the MIT license.


class EmulatorException(Exception):
    def __init__(self, *args):
        super().__init__(*args)


class EmulationExit(EmulatorException):
    def __init__(self):
        super().__init__()
